import audioop
import datetime
import json
import socketserver
import threading
import time
import timeit
from http.server import BaseHTTPRequestHandler, HTTPServer
from secrets import token_urlsafe
from threading import Thread, enumerate
from typing import Tuple, Union

import shared_variables
from utils import get_logger, check_wav_file_format

# Anton imports
import os
import sys
import constants
import traceback
from multiprocessing import Process
from queue import Queue

logger = get_logger("HTTP_Server_backend")


class PagingHttpRequestHandler(BaseHTTPRequestHandler):
    def _set_response(self, code: int = 200, message: str = ""):
        self.send_response(code=code, message=message)
        if message != "":
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps({"response": message}).encode())
        else:
            self.send_header("Content-type", "text/html")
            self.end_headers()

    def do_GET(self):
        # logger.warning("Received a GET request. path = %s", self.path)
        try:
            if self.path == "/clients":
                self._return_active_clients()
            elif self.path == "/groups":
                self._return_groups()
            elif self.path == "/audio_files":
                self._return_audio_files()
            elif self.path == "/users":
                self._return_users()
            else:
                self._set_response(code=404, message="Unknown GET request path")
        except Exception as e:
            logger.warning(f"Exception occurred while executing GET {self.path} request")
            logger.warning(f"{e}")

    def do_PUT(self):
        # logger.warning("Received a PUT request. path = %s", self.path)
        try:
            if self.path == "/paging":
                self._page_audio()
            elif self.path == "/recording":
                self._record_audio()
            elif self.path == "/playfile":
                self._get_playfile_state()
            else:
                self._set_response(code=404, message="Unknown PUT request path")
        except Exception as e:
            logger.warning(f"Exception occurred while executing PUT {self.path} request")
            logger.warning(f"{e}")

    def do_POST(self):
        """
        Post request for starting/stopping paging request.
        """
        # logger.warning("Received a POST request. path = %s", self.path)
        try:
            if self.path == "/paging":
                self._start_stop_page_message()
            elif self.path == "/recording":
                self._start_stop_page_message()
            elif self.path == "/playfile":
                self._start_stop_playing_file()
            elif self.path == "/groups":
                self._create_group()
            elif self.path == "/users":
                self._create_user()
            elif self.path == "/audio_files":
                self._create_audio_file()
            else:
                self._set_response(code=404, message="Unknown POST request path")
        except Exception as e:
            logger.warning(f"Exception occurred while executing POST {self.path} request")
            logger.warning(f"{e}")

    def do_DELETE(self):
        try:
            if self.path == "/groups":
                self._delete_group()
            elif self.path == "/users":
                self._delete_user()
            elif self.path == "/audio_files":
                self._delete_audio_file()
            else:
                self._set_response(code=404, message="Unknown DELETE request path")
        except Exception as e:
            logger.warning(f"Exception occurred while executing DELETE {self.path} request")
            logger.warning(f"{e}")

    def do_UPDATE(self):
        try:
            if self.path == "/groups":
                self._update_group()
            elif self.path == "/audio_files":
                self._update_audio_file()
            elif self.path == "/users":
                self._update_user()
            else:
                self._set_response(code=404, message="Unknown UPDATE request path")
        except Exception as e:
            logger.warning(f"Exception occurred while executing UPDATE {self.path} request")
            logger.warning(f"{e}")

    def _device_idle(self):
        if shared_variables.paging_state or shared_variables.playfile_state:
            # logger.warning("Device is currently being used!")
            return False
        return True

    def _get_playfile_state(self):
        """
            While a file is being played, the front end can get the status by sending a /playfile PUT request
        """
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        if self._check_user(message["token"]):
            if shared_variables.playfile_state == constants.BUSY:
                msg = "Audio file is currently being played"
                response = shared_variables.audio_files.active_audio_file  # Currently used file, if active
                self._set_response(code=206, message=response)
            elif shared_variables.playfile_state == constants.IDLE:
                msg = "No audio file is currently being played"
                logger.info(msg)
                response = shared_variables.audio_files.active_audio_file  # Currently used file, if active
                self._set_response(code=200, message=response)
            else:
                msg = ">>> Unknown playfile state <<<"
                logger.warning(msg)
                self._set_response(code=500, message=msg)
        else:
            logger.warning(f"Internal server error, {message}")
            logger.warning("Unable to authenticate user")
            # The paging state is not ended correctly
            self._set_response(code=500, message="The paging state is not ended correctly")

    # Play a message that has been stored in the memory of the device
    def _start_stop_playing_file(self):
        shared_variables.downloading_recorded_message = constants.NO
        paging_priority = int(shared_variables.config["paging_priority"])
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)

        logger.warning(f"Current message from front end: {message}")

        allow_paging = True
        uuid = None

        try:
            uuid = message['uuid']
        except Exception as e:
            logger.error(f"Following occurred: {e}")

        if self._check_user(message["token"]):
            if uuid is not None:
                allow_paging = self.is_operation_allowed(uuid, "Playing file")

            if not allow_paging:
                logger.warning(f"Same user trying to play a file from a different UUID not allowed!"
                               f"\nRequest -> {message},\nCurrent UUID -> {shared_variables.active_user_uuid}")

                error_msg = "Same user trying to play a file from a different UUID not allowed!"
                self._set_response(code=409, message=error_msg)

            elif not self._device_idle() and message["type"] != 8:
                error_msg = "Operation must be first stopped from all users and devices before proceeding."
                logger.warning(error_msg)
                self._set_response(code=409, message=error_msg)

            # This is the correct order of the messages when playfile should be STOPPED
            elif message["type"] == 8 and shared_variables.last_message_type == 3:
                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.pmaster.send_end_group_request(clients)
                    shared_variables.kill_playfile_thread = constants.YES

                    if shared_variables.playfile_group_thread is not None:  # Make sure no zombie process is left
                        # logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_group_thread")
                        shared_variables.playfile_group_thread.join(timeout=0.5)
                        shared_variables.playfile_group_thread = None
                    if shared_variables.playfile_playing_thread is not None:  # Make sure no zombie process is left
                        # logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_playing_thread")
                        shared_variables.playfile_playing_thread.join(timeout=0.5)
                        shared_variables.playfile_playing_thread = None

                shared_variables.last_client = None
                shared_variables.last_message_type = 8
                shared_variables.active_user = None
                shared_variables.active_user_uuid = None
                shared_variables.playfile_state = constants.IDLE

                # logger.warning("Running threads left")
                # for thread in enumerate():
                #     logger.warning(thread.name)
                logger.info("STOP playing file")
                self._set_response(code=204)

            # This is the correct order of the messages when playfile should be STARTED
            elif message["type"] == 3 and shared_variables.last_message_type == 8:

                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    if shared_variables.playfile_group_thread is None:
                        shared_variables.playfile_group_thread = Thread(
                            target=shared_variables.pmaster.send_page_group_request,
                            daemon=True, name="PlayFileGroup",
                            kwargs={"clients": clients, "priority": paging_priority},
                        )
                        shared_variables.playfile_group_thread.start()
                    else:
                        logger.warning("Thread already running. Needs to be closed first. Returning")
                        return
                    # logger.info("playfile_group_thread started")

                shared_variables.last_message_type = 3
                logger.info("START playing file")

                # The file starts playing from a separate thread
                try:
                    if "button_ID" in message.keys():
                        args = [message["file_ID"], message["button_ID"]]
                    else:
                        logger.warning("Button ID not in request. Using filename...")
                        args = [message["file_ID"], None]
                    shared_variables.playfile_playing_thread = Thread(
                        target=self._play_message, daemon=True, name="PlayFilePlaying",
                        # kwargs={"file_id": message["file_ID"]},
                        args=args
                    )
                    shared_variables.playfile_playing_thread.start()
                    # logger.info("New playfile_playing_thread started as %s",
                    #             playfile_playing_thread.getName())
                except TypeError:
                    logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
                except OSError as error:
                    logger.warning("While in _play_message() the following error occurred:")
                    logger.warning(error)

                # logger.warning("Current running threads")
                # for thread in enumerate():
                #     logger.warning(thread.name)

                # logger.warning(f"Current playfile_group_thread value: {playfile_group_thread}")
                # logger.warning(f"Current playfile_playing_thread value: {playfile_playing_thread}")
                # logger.info("About to start the playing")
                shared_variables.first_fast_packets = 0
                self._set_response(code=204)

            # In case the last message was not ended correctly
            elif message["type"] == 3:
                logger.warning("Last message not ended properly. Ending it first!")
                if shared_variables.last_client:
                    if shared_variables.playfile_group_thread is not None:
                        logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_playing_thread. Not proper ending")
                        shared_variables.playfile_group_thread.join(timeout=0.5)
                        shared_variables.playfile_group_thread = None
                    if shared_variables.playfile_playing_thread is not None:
                        logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_playing_thread. Not proper ending")
                        shared_variables.playfile_playing_thread.join(timeout=0.5)
                        shared_variables.playfile_playing_thread = None

                    shared_variables.pmaster.send_end_group_request(shared_variables.last_client)
                if message["groups"] == "":
                    shared_variables.last_client = None
                    logger.warning("No group selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    shared_variables.playfile_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        daemon=True, name="PlayFileGroup",
                        kwargs={"clients": clients, "priority": paging_priority},
                    )
                    shared_variables.playfile_group_thread.start()
                    logger.info("playfile_group_thread started.")

                shared_variables.last_message_type = 3
                logger.info("START playing file")

                try:
                    if "button_ID" in message.keys():
                        args = [message["file_ID"], message["button_ID"]]
                    else:
                        logger.warning("Button ID not in request. Using filename...")
                        args = [message["file_ID"], None]
                    shared_variables.playfile_playing_thread = Thread(
                        target=self._play_message, daemon=True, name="PlayFilePlaying",
                        # kwargs={"file_id": message["file_ID"]},
                        args=args
                    )
                    shared_variables.playfile_playing_thread.start()
                    logger.info("playfile_playing_thread started")
                except TypeError:
                    logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
                except OSError as error:
                    logger.warning("While in _play_message() the following error occurred:")
                    logger.warning(error)

                shared_variables.first_fast_packets = 0
                self._set_response(code=204)
                shared_variables.playfile_state = constants.IDLE

            else:
                logger.warning(f"Internal server error, {message}")
                self._set_response(code=500, message="The playfile state is not ended correctly")
        else:
            error_msg = f"Another user is playing: {shared_variables.active_user}"
            new_user = shared_variables.users.get_user_from_token(message["token"])
            logger.info(f"Received request from {new_user}")
            logger.warning(error_msg)
            self._set_response(code=409, message=error_msg)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(error_msg).encode())

    def _play_message(self, file_id: str = "", button_id: Union[str, None] = None):
        """
                Used to send the locally stored audio files
            Args:
                file_id: ID (filename) of the file to be played
                button_id: ID of the button pressed
            Returns:
        """

        shared_variables.downloading_recorded_message = constants.NO
        shared_variables.first_fast_packets = 0
        shared_variables.kill_playfile_thread = constants.NO
        audio_file = None

        file_id = check_wav_file_format(file_id)

        files = []  # Store all the files from the given directory
        for file in os.listdir(constants.PROCESSED_FILES_PATH):
            files.append(file)

        # Check if the requested file has a physical equivalent in storage
        if file_id not in files:
            logger.warning("The requested file ID: %s does not exist", file_id)
            shared_variables.playfile_state = constants.IDLE  #
            shared_variables.kill_playfile_thread = constants.YES
            shared_variables.audio_files.set_active_audio_file(None)
        else:
            logger.info(f"File {file_id} found in existing files")
            if button_id:
                shared_variables.audio_files.set_active_audio_button_id(button_id)
            else:
                shared_variables.audio_files.set_active_audio_file(file_id)
            shared_variables.playfile_state = constants.BUSY

        if shared_variables.kill_playfile_thread is not constants.YES:

            file_to_open = os.path.join(constants.PROCESSED_FILES_PATH, file_id)
            try:
                with open(file_to_open, 'rb') as file_in_mem:
                    # TODO: This needs to be properly detected for each file, based on the header data
                    audio_file = file_in_mem.read()[600:]  # skip the riff & wav header and metadata

                # convert the data to 16-bit signed
                audio_file = audioop.ulaw2lin(audio_file, 2)
            except Exception as ex:
                logger.warning("Unable to read the given %s into bytes: %s", file_to_open, ex)
                shared_variables.kill_playfile_thread = constants.YES
            del file_to_open  # Variable not needed anymore

            # Local playback
            try:
                shared_variables.local_playback_process = Thread(
                    target=shared_variables.write_to_local_playback,
                    args=(audio_file,),
                    name="AlsaPlayer"
                )
                shared_variables.local_playback_process.daemon = True
                shared_variables.local_playback_process.start()
                # logger.warning("New local_playback_process started as: %s!", local_playback_process.name)
            except TypeError:
                logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
            except OSError as error:
                logger.warning("While in shared_variables.write_to_local_playback the following error occurred:")
                logger.warning("Check audio device: %s", shared_variables.config["audio_device"]["playback"])
                logger.warning(error)
        else:
            logger.info("Unable to proceed with execution. shared_variables.kill_playfile_thread is set to: %s !!!",
                        shared_variables.kill_playfile_thread)

        # sent to groups
        if (shared_variables.pmaster.state == 130) and \
                audio_file is not None and \
                shared_variables.kill_playfile_thread is not constants.YES:  # MSG 130, send custom command

            data_block_size = 1280
            # delay = float("{0:.4f}".format(data_block_size / 2 / 8 / 1000))
            delay = 0.079

            if shared_variables.config["sample_rate"] == 8000:
                data_block_size = 1280
                # data_block_size = 320
                delay = float("{0:.4f}".format(data_block_size / 2 / 8 / 1000)) - 0.001
                # delay = 0.080
            elif shared_variables.config["sample_rate"] == 24000:
                data_block_size = 1280
                delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"])) - 0.001
                delay = 0.0261  # TODO: This needs to be properly calculated
                # delay = 0.040
            else:
                logger.warning("Sample rate not supported")

            leftover = len(audio_file) % data_block_size
            count = 0

            if shared_variables.config["audio_format"] == "ulaw":
                shared_variables.playfile_state = constants.BUSY
                # shared_variables.audio_files.set_active_audio_file(file_id)
                logger.info("Using %s audio format", shared_variables.config["audio_format"])
                logger.info(f"Sending {data_block_size/2} byte packages with {delay}s delay")

                for i in range(len(audio_file) // data_block_size):
                    # Needed in order to be able to properly kill the thread prematurely
                    if shared_variables.kill_playfile_thread is not constants.YES:
                        if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                            # time.sleep(0.002)
                            time.sleep(delay)  # data_block_size//2//8 ms
                            shared_variables.first_fast_packets += 1
                        else:
                            time.sleep(delay)  # data_block_size//2//8 ms

                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2ulaw(audio_file[i * data_block_size: (i + 1) * data_block_size], 2),
                            (
                                shared_variables.pmaster.page_to_address,
                                shared_variables.pmaster.page_to_port,
                            ),
                        )
                        count = count + data_block_size
                    else:
                        logger.warning("File playing stopped prematurely.")
                        break

                if leftover != 0:
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2ulaw(audio_file[len(audio_file) - leftover:], 2),
                        (
                            shared_variables.pmaster.page_to_address,
                            shared_variables.pmaster.page_to_port,
                        ),
                    )
                    count = count + leftover
                logger.info(f"Sent this many bytes: {count/2}, out of {len(audio_file)/2}")

            elif shared_variables.config["audio_format"] == "alaw":
                shared_variables.playfile_state = constants.BUSY
                # shared_variables.audio_files.set_active_audio_file(file_id)
                logger.info("Using %s audio format", shared_variables.config["audio_format"])

                for i in range(len(audio_file) // data_block_size):
                    # Needed in order to be able to properly stop playing prematurely
                    if shared_variables.kill_playfile_thread is not constants.YES:
                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2alaw(audio_file[i * data_block_size: (i + 1) * data_block_size], 2),
                            (
                                shared_variables.pmaster.page_to_address,
                                shared_variables.pmaster.page_to_port,
                            ),
                        )
                        count = count + data_block_size
                        if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                            # time.sleep(0.002)
                            time.sleep(delay)  # data_block_size//2//8 ms
                            shared_variables.first_fast_packets = shared_variables.first_fast_packets + 1
                        else:
                            time.sleep(delay)  # data_block_size//2//8 ms
                    else:
                        logger.warning("File playing stopped prematurely.")
                        break

                if leftover != 0:
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2alaw(audio_file[len(audio_file) - leftover:], 2),
                        (
                            shared_variables.pmaster.page_to_address,
                            shared_variables.pmaster.page_to_port,
                        ),
                    )
                    count = count + leftover
                logger.info(f"Sent this many bytes: {count/2}, out of {len(audio_file)/2}")

            else:
                logger.warning(
                    f"Audio format not recognized : {shared_variables.config['audio_format']}"
                )
                self._set_response(400)

            # Kill the local playback process. Important for the case when the user stops the playback earlier
            if shared_variables.local_playback_process is not None:
                shared_variables.kill_alsa_player()
                shared_variables.local_playback_process.join(timeout=0.5)
                shared_variables.local_playback_process = None

            shared_variables.playfile_state = constants.IDLE
            shared_variables.audio_files.set_active_audio_file(None)
            logger.info(f"Successfully finished playing file [{file_id}]")

        else:
            # The request could not be processed because of conflict in the current state of the resource,
            self._set_response(409)

    def _record_audio(self):
        # local_playback_process = None

        try:
            content_length = int(
                self.headers["Content-Length"]
            )  # <--- Gets the size of data
        except Exception as e:
            logger.error(f"Error while getting content: {e}")
            return self._set_response(code=409, message=f"Error while getting content: {e}")

        try:
            message = self.rfile.read(content_length)
            token, audio = self._extract_token_and_audio(message)
        except Exception as e:
            logger.error(f"Error while reading the content: {e}")
            return self._set_response(code=409, message=f"Error while reading the content: {e}")

        shared_variables.downloading_recorded_message = constants.YES

        try:
            if token == shared_variables.users.secret(shared_variables.active_user):
                # Local playback
                try:
                    shared_variables.local_playback_process = Thread(
                        target=shared_variables.write_to_local_playback,
                        args=(audio,),
                        name="AlsaPlayer"
                    )
                    shared_variables.local_playback_process.daemon = True
                    shared_variables.local_playback_process.start()
                except TypeError:
                    logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
                except OSError as error:
                    logger.warning("While in shared_variables.local_alsa_player.write() the following error occurred:")
                    logger.warning(error)
                    logger.warning("Check audio device: %s", shared_variables.config["audio_device"]["playback"])

                # sent to groups
                if shared_variables.pmaster.state == 130:
                    data_block_size = 480
                    delay = 0.03
                    logger.info("Sample rate: %s", shared_variables.config["sample_rate"])
                    if shared_variables.config["sample_rate"] == 8000:
                        data_block_size = 912
                        delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                        # delay = 0.02
                    elif shared_variables.config["sample_rate"] == 24000:
                        data_block_size = 964
                        delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                        # delay = 0.02
                    else:
                        logger.warning("Sample rate not supported")

                    # First record the message on the disk
                    with open(shared_variables.record_service.current_file, "ab") as message_file:
                        logger.info(f"Writing audio to: {shared_variables.record_service.current_file}")
                        message_file.write(audio)  # PCM 16bit

                    # Notify that the file has been downloaded
                    shared_variables.downloading_recorded_message = constants.NO
                    time.sleep(0.15)  # Give some time to read the flag
                    # Then read from the recorded file
                    with open(shared_variables.record_service.current_file, "rb") as message_file:
                        audio_file = message_file.read()  # PCM 16bit

                        logger.info("Reading from recorded file...")
                        # logger.info(f"Current fast packets value: {shared_variables.first_fast_packets}")
                        self._encode_send_audio(audio_file, shared_variables.config["audio_format"],
                                                data_block_size, delay)
                else:
                    logger.warning("Current paging master state: %s", shared_variables.pmaster.state)

                # Kill the local playback process. Important for the case when the user stops the playback earlier
                if shared_variables.local_playback_process is not None:
                    shared_variables.kill_alsa_player()
                    shared_variables.local_playback_process.join(timeout=0.5)
                    shared_variables.local_playback_process = None

                self._set_response(code=204)  # Notify front-end we are done
            else:
                msg = f"Could not authenticate user with ID: {token}"
                logger.error(msg)
                self._set_response(code=409, message=msg)

        except KeyError as e:
            # The request could not be processed because of conflict in the current state of the resource,
            if shared_variables.active_user is None:
                msg = f"Key error. The current user is: {e}"
                logger.error(msg)
            else:
                msg = f"Key error. Message: {e}"
                logger.error(msg)
                traceback.print_exc()
            self._set_response(code=409, message=msg)

    def _local_playback_start(self):
        """
            This function is used for the local playback, when live paging is performed.
        Returns:

        """

        if shared_variables.local_playback_msg_queue is None:
            try:
                shared_variables.end_operation = False
                shared_variables.local_playback_msg_queue = Queue(100)  # constants.Q_SIZE
                # logger.info(f"Created local playback queue")
            except Exception as e:
                logger.warning(f"Following occurred: {e}")
        else:
            logger.error("Could not start a new Q. shared_variables.local_playback_msg_queue already exists")

        # Local playback
        if shared_variables.local_playback_queue_process is None:
            try:
                shared_variables.local_playback_queue_process = Thread(
                    target=self._read_from_local_playback_queue,
                    kwargs={"m_queue": shared_variables.local_playback_msg_queue},
                    daemon=True,
                )
                shared_variables.local_playback_queue_process.setName("LocalPlayback")
                # logger.info(f"New local playback thread created")
            except TypeError:
                logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
            except OSError as error:
                logger.warning("Check audio device: %s", shared_variables.config["audio_device"]["playback"])
                logger.warning(error)
        else:
            logger.error("Could not start a new local playback thread. Thread already exists")

    def _local_playback_stop(self):
        # logger.warning("Trying to stop the local playback")
        # Kill the local playback process. Important for the case when the user stops the playback earlier
        shared_variables.end_operation = True
        if shared_variables.local_playback_queue_process is not None:
            shared_variables.local_playback_queue_process.join(timeout=0.5)
            shared_variables.local_playback_queue_process = None
            shared_variables.local_playback_msg_queue = None
        else:
            logger.warning("shared_variables.local_playback_queue_process already None")

        shared_variables.kill_alsa_player()

    def _read_from_local_playback_queue(self, m_queue: Queue):
        """
            Only reading from the queue. Started as a separate thread during live paging
        Args:
            m_queue: The message queue which is written to inside _page_audio
        """

        logger.info(f"Starting reading from local playback queue")
        block_size = 912

        while shared_variables.paging_state and not shared_variables.end_operation:
            try:
                q_size = m_queue.qsize()
                if q_size != 0:
                    data_to_play = m_queue.get(block=False)
                    length = len(data_to_play)
                    leftover = length % block_size
                    count = 0

                    for i in range(length // block_size):
                        audio_to_send = data_to_play[i * block_size: (i + 1) * block_size]
                        shared_variables.write_to_local_playback(data=audio_to_send, live_paging=True)
                        count = count + block_size

                    if leftover != 0:
                        shared_variables.write_to_local_playback(data=data_to_play[length - leftover:],
                                                                 live_paging=True)
                        count = count + leftover
                    # logger.info(f"Sent this many bytes: {count/2}, out of {length/2}")

                    m_queue.task_done()
            except Exception as e:
                logger.warning(f"Following occurred: {e}")
            time.sleep(0.001)

        logger.info("Finished reading from local playback queue")
        # m_queue.task_done()

    def _page_audio(self):
        shared_variables.downloading_recorded_message = constants.NO
        content_length = int(
            self.headers["Content-Length"]
        )  # <--- Gets the size of data
        message = self.rfile.read(content_length)
        token, audio = self._extract_token_and_audio(message)

        try:
            if token == shared_variables.users.secret(shared_variables.active_user):
                # send to groups
                if shared_variables.pmaster.state == 130:
                    shared_variables.paging_message_q.put(audio)  # Add data in the que
                    shared_variables.paging_message_q_buff_msg += 1
                    # logger.info(f"Added to queue {len(audio)}, msgs: {shared_variables.paging_message_q_buff_msg}")

                    # Start the queue reading thread as soon as there is data in the queue. Always 16Kb in the buff.
                    # This way prevent problems with very short button clicks.
                    if shared_variables.paging_message_q_buff_msg == 1:
                        shared_variables.paging_message_q_thread.start()  # This thread is only reading from the queue

                # local playback
                try:
                    shared_variables.local_playback_msg_queue.put(audio)
                except Exception as e:
                    logger.warning(f"Couldn't add to local playback queue: {e}")
                if shared_variables.paging_message_q_buff_msg == 1:
                    # Start reading from the que once there is data
                    shared_variables.local_playback_queue_process.start()

                time.sleep(0.001)  # Give chance to other processes to work [10ms]
                self._set_response(code=204)
            else:
                logger.warning("User token not found!")

        except KeyError as e:
            self._local_playback_stop()
            # The request could not be processed because of conflict in the current state of the resource,
            if shared_variables.active_user is None:
                logger.error(f"Key error. The current user is: {e}")
            else:
                logger.error(f"Key error. Message: {e}")
                traceback.print_exc()
            self._set_response(409)

    def _encode_send_audio_queue(self, m_queue, audio_format="", block_size=1920, delay=0.001):
        """
        Only use this function for live paging
        Reads PCM 16 audio buffers from a message queue, encodes them to either uLaw or aLaw and sends them over
        through the socket to the selected clients.
        Args:
            m_queue: Message queue holding binary IO audio buffers
            audio_format: uLaw or aLaw allowed
            block_size: Packet size to be encoded and sent at a time
            delay: Delay in [s] between the different packets

        Returns: Void
        """

        audio_bin_buff = 0

        # TODO: Debug audio issues after long paging
        # from pathlib import Path
        # live_paging_file = "/tmp/live_paging_test"
        # store_test_file = True
        #
        # file = Path(live_paging_file)
        # file.touch(exist_ok=True)

        # Run until the paging_state is changed to IDLE
        while shared_variables.paging_state:
            if m_queue is not None:
                try:
                    # starttime = timeit.default_timer()
                    audio_bin_buff = m_queue.get(block=True)
                    # endtime = timeit.default_timer()
                    # logger.warning(f"Time to read from queue: {(endtime-starttime)*1000}ms")
                except Exception as e:
                    logger.warning(f"Error occurred while trying to read from queue: {e}")
                q_size = m_queue.qsize()

                count = 0
                length = len(audio_bin_buff)
                leftover = length % block_size

                # logger.info(f"Block size: {block_size}, delay: {delay}, q_len: {q_size}")
                # logger.info(f"Value of first fast packets variable: {shared_variables.first_fast_packets}")

                # logger.info(f"Len of data: {length}")

                if audio_format == "ulaw":
                    # starttime = timeit.default_timer()
                    for i in range(length // block_size):
                        audio_to_send = audio_bin_buff[i * block_size: (i + 1) * block_size]

                        # elapsed_time = time.perf_counter() - self._start_time  // i
                        mulaw_encoded = audioop.lin2ulaw(audio_to_send, 2)
                        shared_variables.pmaster.audio_socket.sendto(
                            mulaw_encoded,
                            (shared_variables.pmaster.page_to_address, shared_variables.pmaster.page_to_port,),
                        )

                        count = count + block_size

                        # Burst the first PACKETS_NUMBER packages to accumulate a buffer to read from,
                        # then go with equal intervals
                        if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                            time.sleep(0.004)
                            shared_variables.first_fast_packets = shared_variables.first_fast_packets + 1
                        else:
                            time.sleep(delay)

                    # If there is data left over from the buffer, send it at the end without delay
                    if leftover != 0:
                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2ulaw(audio_bin_buff[length - leftover:], 2),
                            (shared_variables.pmaster.page_to_address, shared_variables.pmaster.page_to_port,),
                        )
                        count = count + leftover
                    # logger.info(f"Sent this many bytes: {count/2}, out of {length/2}")
                    # endtime = timeit.default_timer()
                    # logger.warning(f"Time needed to send to clients: {(endtime-starttime)*1000}ms")
                elif audio_format == "alaw":
                    for i in range(length // block_size):
                        # elapsed_time = time.perf_counter() - self._start_time  // i
                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2alaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                            (
                                shared_variables.pmaster.page_to_address,
                                shared_variables.pmaster.page_to_port,
                            ),
                        )
                        count = count + block_size

                        # Burst the first PACKETS_NUMBER packages to accumulate a buffer to read from,
                        # then go with equal intervals
                        if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                            time.sleep(0.004)
                            shared_variables.first_fast_packets = shared_variables.first_fast_packets + 1
                        else:
                            time.sleep(delay)

                    # If there is data left over from the buffer, send it at the end without delay
                    if leftover != 0:
                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2alaw(audio_bin_buff[length - leftover:], 2),
                            (
                                shared_variables.pmaster.page_to_address,
                                shared_variables.pmaster.page_to_port,
                            ),
                        )
                        count = count + leftover
                    # logger.info(f"Sent this many bytes: {count / 2}, out of {length / 2}")
                else:
                    msg = f"Audio format not recognized : {audio_format}"
                    logger.warning(msg)
                    self._set_response(code=400, message=msg)
                # logger.info(f"Paging state: {shared_variables.paging_state}")
                m_queue.task_done()
        logger.info("Finished paging thread")

    def _encode_send_audio(self, audio_bin_buff, audio_format="", block_size=1920, delay=0.001):
        """
        Takes the PCM 16 audio buffer, encodes it to either uLaw or aLaw and sends it over through the socket
        to the selected clients
        Args:
            audio_bin_buff: Binary IO audio buffers
            audio_format: uLaw or aLaw allowed
            block_size: Packet size to be encoded and sent at a time
            delay: Delay in [s] between the different packets

        Returns: Void
        """

        count = 0
        length = len(audio_bin_buff)
        leftover = length % block_size

        logger.info(f"Using block size: {block_size}, and delay: {delay}")
        # logger.info(f"Value of first fast packets variable: {shared_variables.first_fast_packets}")

        if audio_format == "ulaw":
            for i in range(length // block_size):
                # elapsed_time = time.perf_counter() - self._start_time  // i
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2ulaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                    (
                        shared_variables.pmaster.page_to_address,
                        shared_variables.pmaster.page_to_port,
                    ),
                )
                count = count + block_size

                if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                    time.sleep(0.002)
                    shared_variables.first_fast_packets += 1
                else:
                    time.sleep(delay)

            if leftover:
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2ulaw(audio_bin_buff[length - leftover:], 2),
                    (
                        shared_variables.pmaster.page_to_address,
                        shared_variables.pmaster.page_to_port,
                    ),
                )
                count = count + leftover
            logger.info(f"Sent this many bytes: {count/2}, out of {length/2}")

        elif audio_format == "alaw":
            for i in range(length // block_size):
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2alaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                    (
                        shared_variables.pmaster.page_to_address,
                        shared_variables.pmaster.page_to_port,
                    ),
                )
                count = count + block_size
                time.sleep(delay)

            if leftover:
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2alaw(audio_bin_buff[length - leftover:], 2),
                    (
                        shared_variables.pmaster.page_to_address,
                        shared_variables.pmaster.page_to_port,
                    ),
                )
                count = count + leftover
            logger.info(f"Sent this many bytes: {count}, out of {len(audio_bin_buff)}")
        else:
            msg = f"Audio format not recognized : {audio_format}"
            logger.warning(msg)
            self._set_response(code=400, message=msg)

    def _check_user(self, secret):
        # Add timer to the user connection in case of bad messages.
        # timer = Timer("_check_user")
        # timer.start()
        if shared_variables.active_user is None:
            shared_variables.active_user = shared_variables.users.get_user_from_token(
                secret
            )
            # timer.stop()
            return True if shared_variables.active_user is not None else False
        elif shared_variables.users.secret(shared_variables.active_user) != secret:
            # timer.stop()
            return False
        # timer.stop()
        return True

    def is_operation_allowed(self, uuid, operation) -> bool:
        msg = f"{operation} is allowed for UUID: {uuid}"
        if shared_variables.active_user_uuid is None:
            shared_variables.active_user_uuid = uuid

        elif uuid != shared_variables.active_user_uuid:
            # TODO: When same user token but different groups are selected, the operations should be allowed
            # Different UUID trying to page from the same user token is not allowed
            msg = f"{operation} NOT allowed for UUID: {uuid}"
            logger.info(msg)
            return False

        logger.info(msg)
        return True

    def _start_stop_page_message(self):
        if not click_allowed():
            front_end_reply = "Button pressed too fast!"
            logger.warning(front_end_reply)
            self._set_response(code=409, message=front_end_reply)
            return
        shared_variables.set_last_click_time(_stamp())

        shared_variables.downloading_recorded_message = constants.YES
        paging_priority = int(shared_variables.config["paging_priority"])
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        logger.info("Message received: %s", message)
        message = json.loads(message)
        allow_paging = True
        uuid = None

        try:
            uuid = message['uuid']
        except Exception as e:
            logger.error(f"Following occurred: {e}")

        if self._check_user(message["token"]):
            # logger.info(f"User matched: {shared_variables.active_user}")

            if uuid is not None:
                allow_paging = self.is_operation_allowed(uuid, "Paging")

            data_block_size = 320
            delay = 0.002
            # logger.info("Sample rate: %s", shared_variables.config["sample_rate"])
            if shared_variables.config["sample_rate"] == 8000:
                data_block_size = 912
                delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                # delay = 0.02
            elif shared_variables.config["sample_rate"] == 24000:
                data_block_size = 964
                delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                # delay = 0.02
            else:
                logger.warning("Sample rate not supported")

            # logger.info("Message type: %s", message["type"])

            if not allow_paging:
                logger.warning(f"Same user trying to page from a different UUID not allowed!"
                               f"\nRequest -> {message},\nCurrent UUID -> {shared_variables.active_user_uuid}")
                error_msg = "Same user trying to page from a different device not allowed!"
                self._set_response(code=409, message=error_msg)

            elif not self._device_idle() and message["type"] != 8:
                error_msg = "Operation must be first stopped from all users and devices before proceeding."
                logger.warning(error_msg)
                self._set_response(code=409, message=error_msg)

            elif message["type"] == 8 and shared_variables.last_message_type == 3:

                if self.path == "/paging":

                    self._local_playback_stop()
                    # Wait until the queue is empty
                    while shared_variables.paging_message_q.empty() is False:
                        logger.warning(f"Waiting for queue to be emptied: {shared_variables.paging_message_q.qsize()}")
                        time.sleep(3)

                    # 16 depends on the settings from the front end.
                    logger.info(f"Total received data: {shared_variables.paging_message_q_buff_msg*16}kb")

                    shared_variables.paging_state = constants.IDLE
                    shared_variables.paging_message_q_buff_msg = 0

                    # First join the queue
                    if shared_variables.paging_message_q is not None:
                        shared_variables.paging_message_q.join()
                        shared_variables.paging_message_q = None

                    # Second join the reading thread
                    if shared_variables.paging_message_q_thread is not None:
                        res = shared_variables.paging_message_q_thread.join(timeout=0.5)
                        shared_variables.paging_message_q_thread = None

                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    if shared_variables.send_page_group_thread is not None:
                        shared_variables.send_page_group_thread.join(timeout=0.5)
                        shared_variables.send_page_group_thread = None
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.pmaster.send_end_group_request(clients)
                shared_variables.last_client = None
                shared_variables.last_message_type = 8
                shared_variables.active_user = None
                shared_variables.active_user_uuid = None
                shared_variables.downloading_recorded_message = constants.YES

                if self.path == "/recording":
                    shared_variables.record_service.exit()

                logger.info("STOP paging")
                self._set_response(code=204)
            elif message["type"] == 3 and shared_variables.last_message_type == 8:

                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    shared_variables.send_page_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        kwargs={"clients": clients, "priority": paging_priority},
                    )
                    shared_variables.send_page_group_thread.daemon = True
                    shared_variables.send_page_group_thread.setName("PageGroup")
                    shared_variables.send_page_group_thread.start()

                    if self.path == "/paging":
                        shared_variables.paging_state = constants.BUSY
                        # TODO: Not tested
                        shared_variables.paging_message_q = Queue(constants.Q_SIZE)
                        shared_variables.paging_message_q_thread = Thread(
                            target=self._encode_send_audio_queue,
                            kwargs={"m_queue": shared_variables.paging_message_q,
                                    "audio_format": shared_variables.config["audio_format"],
                                    "block_size": data_block_size,
                                    "delay": delay},
                            daemon=True,
                        )
                        shared_variables.paging_message_q_thread.setName("PageMsgQ")
                        logger.info("paging_message_q started")

                # Init the message queue for local playback and prep the process
                self._local_playback_start()

                shared_variables.last_message_type = 3
                shared_variables.first_fast_packets = 0
                logger.info("START paging")
                self._set_response(code=204)

            elif message["type"] == 3:
                logger.warning("Last message not ended properly. Ending it first!")

                self._local_playback_stop()
                logger.warning("Current send_page_group_thread value: %s",
                               shared_variables.send_page_group_thread)

                if shared_variables.last_client:
                    if shared_variables.send_page_group_thread is not None:
                        shared_variables.send_page_group_thread.join(timeout=0.5)
                        shared_variables.send_page_group_thread = None
                    shared_variables.pmaster.send_end_group_request(shared_variables.last_client)
                if message["groups"] == "":
                    shared_variables.last_client = None
                    logger.warning("No group selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    shared_variables.send_page_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        kwargs={"clients": clients, "priority": paging_priority},
                    )
                    shared_variables.send_page_group_thread.daemon = True
                    shared_variables.send_page_group_thread.setName("PageGroup")
                    shared_variables.send_page_group_thread.start()

                    if self.path == "/paging":
                        shared_variables.paging_state = constants.BUSY
                        # TODO: Not tested
                        shared_variables.paging_message_q = Queue(constants.Q_SIZE)
                        shared_variables.paging_message_q_thread = Thread(
                            target=self._encode_send_audio_queue,
                            kwargs={"m_queue": shared_variables.paging_message_q,
                                    "audio_format": shared_variables.config["audio_format"],
                                    "block_size": data_block_size,
                                    "delay": delay},
                            daemon=True,
                        )
                        shared_variables.paging_message_q_thread.setName("PageMsgQ")

                logger.info(f"Message q: {shared_variables.local_playback_msg_queue}")

                # Init the message queue for local playback
                if shared_variables.local_playback_msg_queue is None:
                    try:
                        self._local_playback_start()
                    except Exception as e:
                        logger.warning(f"Following occurred: {e}")

                shared_variables.last_message_type = 3
                shared_variables.first_fast_packets = 0
                logger.info("START paging")
                self._set_response(code=204)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps("START paging").encode())

            else:
                logger.warning(f"Internal server error:\nMessage -> {message},\ncurrent state ->")
                # The paging state is not ended correctly
                self._set_response(code=500, message="The paging state is not ended correctly")

        else:
            error_msg = f"Another user is paging: {shared_variables.active_user}"
            new_user = shared_variables.users.get_user_from_token(message["token"])
            logger.info(f"Received request from {new_user}")
            logger.warning(error_msg)
            front_end_reply = "Another user is paging!"
            self._set_response(code=409, message=front_end_reply)

    def _return_active_clients(self):
        self.send_response(code=200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(
            json.dumps({"active_clients": list(shared_variables.pmaster.active_devices.keys())}).encode()
        )

    def _return_users(self):
        try:
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(shared_variables.users.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _return_groups(self):
        try:
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(shared_variables.groups.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _return_audio_files(self):
        try:
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(shared_variables.audio_files.bytes)
            logger.info("Success return audio files")
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _extract_token_and_audio(self, message: bytes) -> Tuple[str, bytes]:
        """
        Extracts the token and audio packet.
        The whole array is assumed to be 16 bit Int array. The token is encoded to its corresponding ASCII code.
        """

        try:
            token = "".join(
                [
                    chr(
                        int.from_bytes(
                            message[2 * i: 2 * i + 2], "little", signed=True
                        )
                    )
                    for i in range(shared_variables.user_token_length)
                ]
            )
            # logger.info(f"The token in audio message is {token}")
        except Exception as e:
            logger.warning(f"Unable to extract the token. Error: {e}")
            token = "-1"
        audio = message[shared_variables.user_token_length*2:]
        # logger.info("Token and audio extracted")
        return token, audio

    def _create_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        # logger.warning("Inside _create_group. Message: %s", message)  # Anton
        try:
            status, reason = shared_variables.groups.add(message)
            if status:
                self.send_response(201)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self._set_response(400)
        except Exception as e:
            logger.warning(f"Create group error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _create_audio_file(self):
        # print("Inside _create_audio_file.")  # Anton
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        logger.info("Message: %s", message)  # Anton
        try:
            status, reason = shared_variables.audio_files.add(message)
            if status:
                self.send_response(201)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self._set_response(400)
        except Exception as e:
            logger.warning(f"Create audio file error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _create_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.add(message)
            if status:
                self.send_response(201)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self.send_response(400)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
        except Exception as e:
            logger.warning(f"Create user error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _update_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.groups.update(message)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            logger.warning(f"Create group error")
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self._set_response(code=500, message=message)

    def _update_audio_file(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.audio_files.update(message)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            msg = f"Update audio file error: {e}"
            logger.error(msg)
            self._set_response(code=500, message=msg)

    def _update_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.update(message)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            msg = f"Create group error: {e}"
            logger.warning(msg)
            self._set_response(code=500, message=msg)

    def _delete_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.groups.remove(message)
            if status:
                self._set_response(200)
            else:
                self._set_response(404)
        except Exception as e:
            msg = f"Delete group error: {e}"
            logger.warning(msg)
            self._set_response(code=500, message=msg)

    def _delete_audio_file(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)

        try:
            status, reason = shared_variables.audio_files.remove(message)
            if status:
                self._set_response(200)
            else:
                self._set_response(404)
        except Exception as e:
            msg = f"Delete audio file error: {e}"
            logger.warning(msg)
            self._set_response(code=500, message=msg)

    def _delete_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.remove(message)
            if status:
                self._set_response(200)
            else:
                self._set_response(404)
        except Exception as e:
            msg = f"Delete user error: {e}"
            logger.warning(msg)
            self._set_response(code=500, message=msg)

    def log_message(self, format, *args) -> None:
        pass


def click_allowed():
    CLICK_TIMEOUT = 1
    """
                Check if the CLICK_TIMEOUT has expired, button can be clicked again
            :return: True if expired or not yet used, False otherwise
            """
    # If default value, the stream is playing normally and no switching has occured
    if shared_variables.last_click_time == -1:
        logger.info("Default case, switching normally")
        return True
    now = _stamp()
    diff = now - shared_variables.last_click_time
    if diff > CLICK_TIMEOUT:
        logger.info(
            f"Timeout reached [{CLICK_TIMEOUT}] -> stream switching can continue normally")
        shared_variables.set_last_click_time(-1)
        return True
    logger.info(f"Wait for timeout - {diff}")
    return False


def _stamp():
    return datetime.datetime.now().timestamp()


httpd = None


# Method for starting and running  the server
def run_server(server_class=HTTPServer, handler_class=PagingHttpRequestHandler, port: int = 8765,):
    global httpd
    server_address = ("0.0.0.0", port)
    httpd = server_class(server_address, handler_class)
    logger.info("Starting httpd...")
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass

    if httpd is not None:
        httpd.server_close()
    logger.info("Stopping httpd...")


def stop_server():
    # Kill the local playback process. Important for the case when the user stops the playback earlier
    if shared_variables.local_playback_process is not None:
        shared_variables.kill_alsa_player()
        shared_variables.local_playback_process.join(timeout=0.5)
        shared_variables.local_playback_process = None

    shared_variables.playfile_state = constants.IDLE
    shared_variables.audio_files.set_active_audio_file(None)

    global httpd
    if httpd is not None:
        httpd.server_close()
        logger.info("Stopping httpd...")
